import click
from datetime import datetime

from .__init__ import jakt
from .timeslot import timeslot
from .exceptions import *


@click.group()
@click.version_option(version="0.0.4", prog_name="jakt")
@click.pass_context
def cli(ctx):
    """Jakt is just another (k)ommandline timetracker.

    Jakt helps you keep track of how you spend your time.
    Whether you want to keep better track of how much time
    you spend on each project or want to keep yourself
    accountable while working, jakt is the perfect tool."""

    ctx.ensure_object(dict)
    ctx.obj["jakt"] = jakt()


@cli.command()
@click.argument("project")
@click.argument("tags", nargs=-1)
@click.pass_context
def start(ctx, project, tags):
    """Start a new timeslot"""
    jkt = ctx.obj["jakt"]

    try:
        response = jkt.start(project=project, tags=tags)

        project = click.style(response.project, fg="blue", bold=True)
        hrStart = click.style(
            datetime.fromtimestamp(response.start).strftime("%H:%M"),
            fg="red",
            bold=True,
        )
        tags = click.style(" ".join(str(t) for t in response.tags), fg="green")

        click.echo(f"{project} started at {hrStart}")
        click.echo(f"Tags: {tags}")

    except JaktActiveError:
        click.echo("Other timer already running")
        ctx.invoke(status)


@cli.command()
@click.pass_context
def stop(ctx):
    """Stops current project"""
    jkt = ctx.obj["jakt"]

    try:
        ts = jkt.stop()

        project = click.style(ts.project, fg="blue", bold=True)
        tags = click.style(" ".join(str(t) for t in ts.tags), fg="green")

        hrStop = click.style(ts.end_dt.strftime("%H:%M"), fg="red", bold=True)
        dur = ts.getDurationHR()
        runtime = click.style(f"{dur['hh']:02}:{dur['M']:02}", fg="green", bold=True)

        click.echo(f"{project} stopped at {hrStop}")
        click.echo(f"Tags: {tags}")
        click.echo(f"Timer ran for {runtime}")

    except JaktNotActiveError:
        click.echo("No timer started.")


@cli.command()
@click.pass_context
def status(ctx):
    """Displays current status"""
    jkt = ctx.obj["jakt"]

    try:
        response = jkt.status()

        project = click.style(response["project"], fg="blue", bold=True)
        hrStart = click.style(
            datetime.fromtimestamp(response["start"]).strftime("%H:%M"),
            fg="red",
            bold=True,
        )
        tags = click.style(" ".join(str(t) for t in response["tags"]), fg="green")
        runtime = click.style(
            f"{response['elapsedHour']:02}:{response['elapsedMin']:02}",
            fg="green",
            bold=True,
        )

        click.echo(f"{project} started at {hrStart}.")
        click.echo(f"Tags: {tags}")
        click.echo(f"Runtime is {runtime}")

    except JaktNotActiveError:
        click.echo("No timer started.")


@cli.command()
@click.option(
    "--to",
    "to",
    type=click.DateTime(formats=["%d-%m-%y"]),
    help="Starttime of search period",
)
@click.option(
    "--from",
    "from_",
    type=click.DateTime(formats=["%d-%m-%y"]),
    help="Endtime of search period",
)
@click.option(
    "-c", "--categories", is_flag=True, default=False, help="Display categories"
)
@click.option("-p", "--projects", is_flag=True, default=False, help="Display projects")
@click.option("-t", "--tags", is_flag=True, default=False, help="Display tags")
# TODO: Display all timeslots using -a
@click.pass_context
def ls(ctx, to, from_, categories, projects, tags):
    """Lists timeslots and other data"""
    jkt = ctx.obj["jakt"]

    if categories:
        categories = jkt.getCategories()

        for cat in categories:
            click.echo(f"{click.style(cat, fg='red', bold=True)}")

        return

    if projects:
        projects = jkt.getProjects()

        for project in projects:
            click.echo(f"{click.style(project, fg='blue', bold=True)}")

        return

    if tags:
        tags = jkt.getTags()

        for tag in tags:
            click.echo(f"{click.style(tag, fg='green')}")

        return

    if from_ or to:
        # A few cases of input sanitation
        if from_ and (not to):
            to = datetime.now()

        elif to and (not from_):
            click.echo("--from must be set if --to is set")
            return

        if from_ > to:
            # Switch the parameters if they are given in the wrong order
            a = from_
            from_ = to
            to = a
            click.echo("--to/--from in wrong order. Flipping them.")

        timeslots = jkt.getTimeslots(to, from_)
    else:
        timeslots = jkt.getTimeslots()

    # Want timeslots sorted chronologically
    timeslots.reverse()

    for i in range(len(timeslots)):
        if i == 10:
            click.echo(f"{len(timeslots)-i} timeslots not shown.")
            break

        ts = timeslots[i]

        # Define all styles and shown data
        ts_id = click.style(ts.id, fg="yellow")
        ts_project = click.style(ts.project, fg="blue", bold=True)

        # Make sure time is readable and makes sense
        ts_start = ts.start_dt
        ts_end = ts.end_dt

        if ts_start.date() == ts_end.date():
            ts_start_hr = ts_start.strftime("%H:%M")
        else:
            ts_start_hr = ts_start.strftime("%H:%M %d-%m-%y")

        ts_end_hr = ts_end.strftime("%H:%M %d-%m-%y")

        s = str(ts.duration).split(":")
        ts_duration = click.style(
            f"{int(s[0]):02}:{int(s[1]):02}:{int(s[2]):02}", fg="green"
        )

        ts_tags = click.style(" ".join(str(t) for t in ts.tags), fg="green")

        # Display data on single line
        click.echo(
            f"{ts_id} {ts_duration} ({ts_start_hr} - {ts_end_hr}) {ts_project} {ts_tags}"
        )


@cli.command()
@click.option(
    "--to",
    "to",
    type=click.DateTime(formats=["%d-%m-%y %H:%M", "%d-%m-%y %H:%M:%S"]),
    help="Starttime",
    required=True,
)
@click.option(
    "--from",
    "from_",
    type=click.DateTime(formats=["%d-%m-%y %H:%M", "%d-%m-%y %H:%M:%S"]),
    help="Endtime",
    required=True,
)
@click.argument("project")
@click.argument("tags", nargs=-1)
@click.pass_context
def add(ctx, to, from_, project, tags):
    """Add a timeslot that was not logged live"""
    jkt = ctx.obj["jakt"]

    ts = timeslot(
        ID=jkt.generateUniqueID(),
        start=int(from_.strftime("%s")),
        end=int(to.strftime("%s")),
        project=project,
        tags=tags,
    )

    jkt.add(ts)


"""
@cli.command()
@click.argument("index")
def edit(index):
    # Edits categories, projects, tags and timeslots
    pass
"""


@cli.command()
@click.option("-p", "--project", default="", help="Show only specified project")
@click.option("-t", "--tag", default="", help="Show only specified tag")
@click.pass_context
def report(ctx, project, tag):
    """Generates reports from timetracker data"""
    jkt = ctx.obj["jakt"]

    jkt_report = jkt.report()

    if project:
        projects = jkt_report.getProjectReport(project=project)
    else:
        projects = jkt_report.getProjectReport()

    for project in projects:
        hrProject = click.style(project["project"], fg="blue", bold=True)
        hrTime = click.style(project["time"], fg="red", bold=True)
        click.echo(f"{hrProject}  {hrTime}")

        tags = jkt_report.getTagReport(project["project"])
        for tag in tags:
            hrTag = click.style(tag["tag"], fg="green", bold=True)
            hrTagTime = click.style(tag["time"], fg="yellow")

            click.echo(f" - {hrTag}  {hrTagTime}")



@cli.command()
@click.pass_context
def resume(ctx):
    """
    Start new timeslot with same settings
    """
    jkt = ctx.obj["jakt"]

    try:
        response = jkt.resume()

        project = click.style(response.project, fg="blue", bold=True)
        hrStart = click.style(
            datetime.fromtimestamp(response.start).strftime("%H:%M"),
            fg="red",
            bold=True,
        )
        tags = click.style(" ".join(str(t) for t in response.tags), fg="green")

        click.echo(f"{project} started at {hrStart}")
        click.echo(f"Tags: {tags}")

    except JaktActiveError:
        click.echo("Other timer already running")
        ctx.invoke(status)


"""
@cli.command()
def pause():
    # Takes a break in current timeslot
    pass

@cli.command()
def config():
    # Sets new values in configuration file
    pass


@cli.command()
def sync():
    # Syncronizes data with server
    pass

@cli.command()
def license():
    # Outputs license
    pass
"""

if __name__ == "__main__":
    cli(obj={})



# Selfdefined Errors
class JaktError(Exception):
    pass


class JaktActiveError(JaktError):
    pass


class JaktNotActiveError(JaktError):
    pass


class JaktPathError(JaktError):
    def __init__(self, path):
        self.path = path


class JaktInputError(JaktError):
    pass



import os
import yaml
import json
import random
from datetime import datetime
from time import time
import click

from .timeslot import timeslot
from .report import JaktReport
from .exceptions import *


class jakt:
    def __init__(self):
        # TODO: Set up config
        # TODO: Read from config path and set variables

        self.dataPath = os.path.join(os.path.expanduser("~"), ".jakt")
        self.pathConfig = os.path.join(self.dataPath, "config.yml")
        self.pathCategories = os.path.join(self.dataPath, "categories.yml")
        self.pathProjects = os.path.join(self.dataPath, "projects.json")
        self.pathTimeslots = os.path.join(self.dataPath, "timeslots.json")
        self.pathCurrent = os.path.join(self.dataPath, "current.json")

        # Standard setup for first time use.
        if not os.path.exists(self.dataPath):
            os.mkdir(self.dataPath)

            # Create standard files
            paths = [
                self.pathConfig,
                self.pathCategories,
                self.pathProjects,
                self.pathTimeslots,
            ]

            for path in paths:
                # Create standard config
                f = open(path, "x")
                f.close()

            standardConfig = {"Remote": False}
            with open(self.pathConfig, "a") as f:
                yaml.dump(standardConfig, f, default_flow_style=False)
        else:
            # TODO: Set up config-loading
            pass

    ## Main working functions
    def start(self, project: str, tags: list[str]) -> dict:
        """
        Adds inputed data into the current file in jakt directory.
        """

        if os.path.exists(self.pathCurrent):
            raise JaktActiveError

        if tags == ():
            tags = ["<no tags>"]

        ts = timeslot(
            ID=self.generateUniqueID(),
            start=round(time()),
            end=None,
            project=project,
            tags=tags
        )

        with open(self.pathCurrent, "w") as f:
            f.write(str(ts.toDictString()))
            f.close()

        return ts

    def stop(self) -> timeslot:
        if not os.path.exists(self.pathCurrent):
            raise JaktNotActiveError

        # Update status from file
        self.status()

        #  Create object to add
        ts = timeslot(
            ID=self.generateUniqueID(),
            start=self.activeTimeslot["start"],
            end=round(time()),
            project=self.activeTimeslot["project"],
            tags=self.activeTimeslot["tags"],
        )

        # Add object to timeslots
        ts_added = self.add(ts)

        # Removes timeslot data in current timeslot
        os.remove(self.pathCurrent)

        return ts_added

    def status(self) -> dict:
        if not os.path.exists(self.pathCurrent):
            raise JaktNotActiveError

        with open(self.pathCurrent, "r") as f:
            status = json.load(f)
            f.close()

        elapsedTime = datetime.fromtimestamp(round(time())) - datetime.fromtimestamp(
            status["start"]
        )
        status["elapsed"] = elapsedTime.seconds

        hours, remainder = divmod(elapsedTime.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        status["elapsedHour"] = hours
        if seconds > 30:
            status["elapsedMin"] = minutes + 1
        else:
            status["elapsedMin"] = minutes

        self.activeTimeslot = status

        return status

    def add(self, ts: timeslot) -> timeslot:
        """
        Adds new timeslot from.
        TODO: Implement add for known data.
        """

        # Find all logged timeslots
        timeslots = self.getTimeslots()

        # Append new timeslot to list
        timeslots.append(ts)

        # Write all timeslots, including newly added to file
        self.putTimeslots(timeslots)

        return ts

    def report(self) -> JaktReport:
        """
        Returns a JaktReport object
        """

        return JaktReport(self)

    def resume(self) -> timeslot:
    	"""
    	Starts new timeslot with same options as previously logged timeslot
    	"""

    	# Get last logged timeslot
    	timeslots = self.getTimeslots()
    	timeslots.reverse()
    	last_ts = timeslots[0]

    	response = self.start(project = last_ts.project, tags = last_ts.tags)

    	return response


    ## Get and put data
    def getCategories(self) -> list[str]:
        """
        Returns a list of all defined categories
        """
        try:
            with open(self.pathCategories, "r") as f:
                categories = yaml.safe_load(f)
                f.close()

            return categories
        except OSError:
            raise JaktPathError(self.pathCategories)

    def getProjects(self) -> list[str]:
        """
        Returns list of all projects
        """
        timeslots = self.getTimeslots()

        projects = []
        for i in range(len(timeslots)):
            if timeslots[i].project not in projects:
                projects.append(timeslots[i].project)

        return projects

    def getTags(self, project: str = None) -> list[str]:
        """
        Returns a list of all used tags.

        If project is given only tags for the matching project are given.
        """
        timeslots = self.getTimeslots()

        tags = []
        for i in range(len(timeslots)):
            if project and not (project == timeslots[i].project):
                continue

            currentTags = timeslots[i].tags

            for j in range(len(currentTags)):
                if currentTags[j] not in tags:
                    tags.append(currentTags[j])

        return tags

    def getTimeslots(
        self, from_=False, to=False, project=False, tag=False
    ) -> list[timeslot]:
        """
        Returns a list of logged timeslots
        """
        try:
            with open(self.pathTimeslots, "r") as f:
                timeslots = json.load(f)
                f.close()

            # Create timeslot instances
            for i in range(len(timeslots)):
                timeslots[i] = timeslot.from_json(timeslots[i])

            # TODO: Implement filtering with to and from_
            if to and from_:
                # Remove timeslots that do not match
                pass

            # Filters by project if project is given
            if project:
                project_filter = []
                for ts in timeslots:
                    if ts.project == project:
                        project_filter.append(ts)

                timeslots = project_filter

            # Filters by tags if tage are given
            if tag:
                tag_filter = []
                for ts in timeslots:
                    if tag in ts.tags:
                        tag_filter.append(ts)

                timeslots = tag_filter

            return timeslots
        except OSError:
            raise JaktPathError(self.pathTimeslots)

    def getTimeslot(self, queryId: str) -> timeslot:
        timeslots = self.getTimeslots()

        for ts in timeslots:
            if ts.id == queryId:
                return ts

        return False

    def putTimeslots(self, timeslots: list[timeslot]):
        obj_list = []

        for ts in timeslots:
            obj_list.append(ts.toDict())

        try:
            with open(self.pathTimeslots, "w") as f:
                json.dump(obj_list, f)
                f.close()
        except OSError:
            raise JaktPathError(self.pathTimeslots)

    ## Helper functions
    def generateUniqueID(self):
        timeslots = self.getTimeslots()

        usedIDs = []
        for ts in timeslots:
            usedIDs.append(ts.id)

        ID = "%08x" % random.randrange(16**8)
        if ID not in usedIDs:
            return ID
        else:
            return self.generateUniqueID()

    ## Remote syncronization
    def fetch(self):
        pass

    def pull(self):
        pass

    def push(self):
        pass



from datetime import datetime, timedelta
import click


class JaktReport:
    def __init__(self, jkt):
        # self.categories = jkt.getCategories()
        projects = jkt.getProjects()

        self.data = []
        for i in range(len(projects)):
            project = projects[i]

            projectDuration = timedelta(0)

            tags = jkt.getTags(project=project)

            for j in range(len(tags)):
                timeslots = jkt.getTimeslots(project=project, tag=tags[j])

                tagDuration = timedelta(0)
                for ts in timeslots:
                    tagDuration += ts.duration

                tagobj = {"tag": tags[j], "timeslots": timeslots, "time": tagDuration}

                tags[j] = tagobj

            # Calculate project duration
            timeslots = jkt.getTimeslots(project=project)
            for ts in timeslots:
                projectDuration += ts.duration

            projectObj = {"project": project, "tags": tags, "time": projectDuration}

            self.data.append(projectObj)

    def __str__(self):
        return f"{self.data}"

    def hrDuration(self, td):
        s = str(td).split(":")
        return f"{int(s[0]):02}:{int(s[1]):02}:{int(s[2]):02}"

    def getProjectReport(self, project: str = "") -> list[dict]:
        report = []
        for proj in self.data:
            proj_report = {
                "project": proj["project"],
                "time": self.hrDuration(proj["time"]),
            }

            if project:
                if proj["project"] == project:
                    report.append(proj_report)
                continue
            report.append(proj_report)

        return report

    def getTagReport(self, project: str = "") -> list[dict]:
        """
        Returns a report of all tags for a given project
        """
        if not project:
            raise JaktInputError

        # Find project that matches given project string
        for proj in self.data:
            if proj["project"] == project:
                selectedProject = proj

        # Generate report
        report = []
        for tg in selectedProject["tags"]:
            tag_report = {"tag": tg["tag"], "time": self.hrDuration(tg["time"])}
            report.append(tag_report)

        return report



from datetime import datetime
import json


class timeslot:
    def __init__(self, ID: str, start: int, end: int, project: str, tags: list[str]):
        self.id = ID

        self.start = start
        self.end = end
        self.start_dt = datetime.fromtimestamp(self.start)
        if end:
            self.end_dt = datetime.fromtimestamp(self.end)

            start = datetime.fromtimestamp(self.start)
            end = datetime.fromtimestamp(self.end)
            self.duration = end - start

        else:
            self.end_dt = None
            self.duration = None

        self.project = project
        self.tags = tags
        

    def __str__(self):
        return f"ts: {self.id} {self.project} {self.tags} {self.start_dt.strftime('%d-%m-%y %H:%M')} - {self.end_dt.strftime('%H:%M')}"

    @classmethod
    def from_json(cls, json_obj: dict):
        """
        Initialize a timeslot directly from a dictionary object.
        """
        return cls(
            ID=json_obj["id"],
            start=json_obj["start"],
            end=json_obj["end"],
            project=json_obj["project"],
            tags=json_obj["tags"],
        )

    def toDict(self) -> dict:
        """
        Returns a dictionary of the timeslot
        """
        obj = {
            "id": self.id,
            "start": self.start,
            "end": self.end,
            "project": self.project,
            "tags": self.tags,
        }

        return obj

    def toDictString(self) -> str:
        """
        Returns a JSON-string
        """

        return json.dumps(self.toDict())

    def getDurationHR(self):
        """
        Return the data needed to display Human Readale duration
        """
        hh, remainder = divmod(int(self.duration.total_seconds()), 3600)
        mm, ss = divmod(remainder, 60)

        M = mm
        if ss > 30:
            M = mm + 1

        return {"hh": hh, "H": hh, "mm": mm, "M": M, "ss": ss}

    def toHR(self):
        """
        Returns timeslot in interface friendly format
        """

        tags_hr = " ".join(str(t) for t in self.tags)

        # Make sure time is readable and makes sense
        if self.start_dt.date() == self.end_dt.date():
            start_hr = self.start_dt.strftime("%H:%M")
        else:
            start_hr = self.start_dt.strftime("%H:%M %d-%m-%y")

        end_hr = self.end_dt.strftime("%H:%M %d-%m-%y")

        returnString = f"{self.id} {self.duration} ({start_hr} - {end_hr}) {self.project} {tags_hr}"

        return returnString
